import string

import hypothesis.strategies as st
import pytest
from hypothesis import given, settings

from tests.utils import decimal_to_int
from vyper.compiler import compile_code
from vyper.exceptions import ArgumentException, CallViolation


def test_selfcall_code(get_contract):
    selfcall_code = """
@internal
def _foo() -> int128:
    return 3

@external
def bar() -> int128:
    return self._foo()
    """

    c = get_contract(selfcall_code)
    assert c.bar() == 3

    print("Passed no-argument self-call test")


def test_selfcall_code_2(get_contract, keccak):
    selfcall_code_2 = """
@internal
def _double(x: int128) -> int128:
    return x * 2

@external
def returnten() -> int128:
    return self._double(5)

@internal
def _hashy(x: bytes32) -> bytes32:
    return keccak256(x)

@external
def return_hash_of_rzpadded_cow() -> bytes32:
    return self._hashy(0x636f770000000000000000000000000000000000000000000000000000000000)
    """

    c = get_contract(selfcall_code_2)
    assert c.returnten() == 10
    assert c.return_hash_of_rzpadded_cow() == keccak(b"cow" + b"\x00" * 29)

    print("Passed single fixed-size argument self-call test")


# test that side-effecting self calls do not get optimized out
def test_selfcall_optimizer(get_contract):
    code = """
counter: uint256

@internal
def increment_counter() -> uint256:
    self.counter += 1
    return self.counter
@external
def foo() -> (uint256, uint256):
    x: uint256 = unsafe_mul(self.increment_counter(), 0)
    return x, self.counter
    """
    c = get_contract(code)
    assert c.foo() == (0, 1)


def test_selfcall_code_3(get_contract, keccak):
    selfcall_code_3 = """
@internal
def _hashy2(x: Bytes[100]) -> bytes32:
    return keccak256(x)

@external
def return_hash_of_cow_x_30() -> bytes32:
    return self._hashy2(b"cowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcowcow")  # noqa: E501

@internal
def _len(x: Bytes[100]) -> uint256:
    return len(x)

@external
def returnten() -> uint256:
    return self._len(b"badminton!")
    """

    c = get_contract(selfcall_code_3)
    assert c.return_hash_of_cow_x_30() == keccak(b"cow" * 30)
    assert c.returnten() == 10

    print("Passed single variable-size argument self-call test")


def test_selfcall_code_4(get_contract):
    selfcall_code_4 = """
@internal
def _summy(x: int128, y: int128) -> int128:
    return x + y

@internal
def _catty(x: Bytes[5], y: Bytes[5]) -> Bytes[10]:
    return concat(x, y)

@internal
def _slicey1(x: Bytes[10], y: uint256) -> Bytes[10]:
    return slice(x, 0, y)

@internal
def _slicey2(y: uint256, x: Bytes[10]) -> Bytes[10]:
    return slice(x, 0, y)

@external
def returnten() -> int128:
    return self._summy(3, 7)

@external
def return_mongoose() -> Bytes[10]:
    return self._catty(b"mon", b"goose")

@external
def return_goose() -> Bytes[10]:
    return self._slicey1(b"goosedog", 5)

@external
def return_goose2() -> Bytes[10]:
    return self._slicey2(5, b"goosedog")
    """

    c = get_contract(selfcall_code_4)
    assert c.returnten() == 10
    assert c.return_mongoose() == b"mongoose"
    assert c.return_goose() == b"goose"
    assert c.return_goose2() == b"goose"

    print("Passed multi-argument self-call test")


def test_selfcall_code_5(get_contract):
    selfcall_code_5 = """
counter: int128

@internal
def _increment():
    self.counter += 1

@external
def returnten() -> int128:
    for i: uint256 in range(10):
        self._increment()
    return self.counter
    """
    c = get_contract(selfcall_code_5)
    assert c.returnten() == 10

    print("Passed self-call statement test")


def test_selfcall_code_6(get_contract):
    selfcall_code_6 = """
excls: Bytes[32]

@internal
def _set_excls(arg: Bytes[32]):
    self.excls = arg

@internal
def _underscore() -> Bytes[1]:
    return b"_"

@internal
def _hardtest(x: Bytes[100], y: uint256, z: uint256, a: Bytes[100], b: uint256, c: uint256) -> Bytes[201]:  # noqa: E501
    return concat(slice(x, y, z), self._underscore(), slice(a, b, c))

@external
def return_mongoose_revolution_32_excls() -> Bytes[201]:
    self._set_excls(b"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    return self._hardtest(b"megamongoose123", 4, 8, concat(b"russian revolution", self.excls), 8, 42)
    """

    c = get_contract(selfcall_code_6)
    assert c.return_mongoose_revolution_32_excls() == b"mongoose_revolution" + b"!" * 32

    print("Passed composite self-call test")


def test_list_call(get_contract):
    code = """
@internal
def _foo0(x: int128[2]) -> int128:
    return x[0]

@internal
def _foo1(x: int128[2]) -> int128:
    return x[1]


@external
def foo1(x: int128[2]) -> int128:
    return self._foo1(x)

@external
def bar() -> int128:
    x: int128[2] = [0, 0]
    return self._foo0(x)

@external
def bar2() -> int128:
    x: int128[2] = [55, 66]
    return self._foo0(x)

@external
def bar3() -> int128:
    x: int128[2] = [55, 66]
    return self._foo1(x)
    """

    c = get_contract(code)
    assert c.bar() == 0
    assert c.foo1([0, 0]) == 0
    assert c.bar2() == 55
    assert c.bar3() == 66


def test_list_storage_call(get_contract):
    code = """
y: int128[2]

@internal
def _foo0(x: int128[2]) -> int128:
    return x[0]

@internal
def _foo1(x: int128[2]) -> int128:
    return x[1]

@external
def set():
    self.y  = [88, 99]

@external
def bar0() -> int128:
    return self._foo0(self.y)

@external
def bar1() -> int128:
    return self._foo1(self.y)
    """

    c = get_contract(code)
    c.set()
    assert c.bar0() == 88
    assert c.bar1() == 99


def test_multi_arg_list_call(get_contract):
    code = """
@internal
def _foo0(y: decimal, x: int128[2]) -> int128:
    return x[0]

@internal
def _foo1(x: int128[2], y: decimal) -> int128:
    return x[1]

@internal
def _foo2(y: decimal, x: int128[2]) -> decimal:
    return y

@internal
def _foo3(x: int128[2], y: decimal) -> int128:
    return x[0]

@internal
def _foo4(x: int128[2], y: int128[2]) -> int128:
    return y[0]


@external
def foo1(x: int128[2], y: decimal) -> int128:
    return self._foo1(x, y)

@external
def bar() -> int128:
    x: int128[2] = [0, 0]
    return self._foo0(0.3434, x)

# list as second parameter
@external
def bar2() -> int128:
    x: int128[2] = [55, 66]
    return self._foo0(0.01, x)

@external
def bar3() -> decimal:
    x: int128[2] = [88, 77]
    return self._foo2(1.33, x)

# list as first parameter
@external
def bar4() -> int128:
    x: int128[2] = [88, 77]
    return self._foo1(x, 1.33)

@external
def bar5() -> int128:
    x: int128[2] = [88, 77]
    return self._foo3(x, 1.33)

# two lists
@external
def bar6() -> int128:
    x: int128[2] = [88, 77]
    y: int128[2] = [99, 66]
    return self._foo4(x, y)

    """

    c = get_contract(code)
    assert c.bar() == 0
    assert c.foo1([0, 0], decimal_to_int("0")) == 0
    assert c.bar2() == 55
    assert c.bar3() == decimal_to_int("1.33")
    assert c.bar4() == 77
    assert c.bar5() == 88


def test_multi_mixed_arg_list_call(get_contract):
    code = """
@internal
def _fooz(x: int128[2], y: decimal, z: int128[2], a: decimal) -> int128:
    return z[1]

@internal
def _fooa(x: int128[2], y: decimal, z: int128[2], a: decimal) -> decimal:
    return a

@external
def bar() -> (int128, decimal):
    x: int128[2] = [33, 44]
    y: decimal = 55.44
    z: int128[2] = [55, 66]
    a: decimal = 66.77

    return self._fooz(x, y, z, a), self._fooa(x, y, z, a)
    """
    c = get_contract(code)
    assert c.bar() == (66, decimal_to_int("66.77"))


def test_internal_function_multiple_lists_as_args(get_contract):
    code = """
@internal
def _foo(y: int128[2], x: Bytes[5]) -> int128:
    return y[0]

@internal
def _foo2(x: Bytes[5], y: int128[2]) -> int128:
    return y[0]

@external
def bar() -> int128:
    return self._foo([1, 2], b"hello")

@external
def bar2() -> int128:
    return self._foo2(b"hello", [1, 2])
"""

    c = get_contract(code)
    assert c.bar() == 1
    assert c.bar2() == 1


def test_multi_mixed_arg_list_bytes_call(get_contract):
    code = """
@internal
def _fooz(x: int128[2], y: decimal, z: Bytes[11], a: decimal) -> Bytes[11]:
    return z

@internal
def _fooa(x: int128[2], y: decimal, z: Bytes[11], a: decimal) -> decimal:
    return a

@internal
def _foox(x: int128[2], y: decimal, z: Bytes[11], a: decimal) -> int128:
    return x[1]


@external
def bar() -> (Bytes[11], decimal, int128):
    x: int128[2] = [33, 44]
    y: decimal = 55.44
    z: Bytes[11] = b"hello world"
    a: decimal = 66.77

    return self._fooz(x, y, z, a), self._fooa(x, y, z, a), self._foox(x, y, z, a)
    """
    c = get_contract(code)
    assert c.bar() == (b"hello world", decimal_to_int("66.77"), 44)


FAILING_CONTRACTS_CALL_VIOLATION = [
    """
# should not compile - public to public
@external
def bar() -> int128:
    return 1

@external
def foo() -> int128:
    return self.bar()
    """,
    """
# should not compile - internal to external
@external
def bar() -> int128:
    return 1

@internal
def _baz() -> int128:
    return self.bar()

@external
def foo() -> int128:
    return self._baz()
    """,
]


@pytest.mark.parametrize("failing_contract_code", FAILING_CONTRACTS_CALL_VIOLATION)
def test_selfcall_call_violation(failing_contract_code, assert_compile_failed):
    with pytest.raises(CallViolation):
        _ = compile_code(failing_contract_code)


FAILING_CONTRACTS_ARGUMENT_EXCEPTION = [
    """
# expected no args, args given
@internal
def bar() -> int128:
    return 1

@external
def foo() -> int128:
    return self.bar(1)
    """,
    """
# expected args, none given
@internal
def bar(a: int128) -> int128:
    return 1

@external
def foo() -> int128:
    return self.bar()
    """,
    """
# wrong arg count
@internal
def bar(a: int128) -> int128:
    return 1

@external
def foo() -> int128:
    return self.bar(1, 2)
    """,
    """
@internal
def _foo(x: uint256, y: uint256 = 1):
    pass

@external
def foo(x: uint256, y: uint256):
    self._foo(x, y=y)
    """,
]


@pytest.mark.parametrize("failing_contract_code", FAILING_CONTRACTS_ARGUMENT_EXCEPTION)
def test_selfcall_wrong_arg_count(failing_contract_code, assert_compile_failed):
    assert_compile_failed(lambda: compile_code(failing_contract_code), ArgumentException)


FAILING_CONTRACTS_TYPE_MISMATCH = [
    """
# should not compile - value kwarg when calling {0} function
@{0}
def foo():
    pass

@external
def bar():
    self.foo(value=100)
    """,
    """
# should not compile - gas kwarg when calling {0} function
@{0}
def foo():
    pass

@external
def bar():
    self.foo(gas=100)
    """,
    """
# should not compile - arbitrary kwargs when calling {0} function
@{0}
def foo():
    pass

@external
def bar():
    self.foo(baz=100)
    """,
    """
# should not compile - args-as-kwargs to a {0} function
@{0}
def foo(baz: int128):
    pass

@external
def bar():
    self.foo(baz=100)
    """,
]


@pytest.mark.parametrize("failing_contract_code", FAILING_CONTRACTS_TYPE_MISMATCH)
@pytest.mark.parametrize("decorator", ["external", "internal"])
def test_selfcall_kwarg_raises(failing_contract_code, decorator, assert_compile_failed):
    exc = ArgumentException if decorator == "internal" else CallViolation
    with pytest.raises(exc):
        _ = compile_code(failing_contract_code.format(decorator))


@pytest.mark.parametrize("i,ln,s,", [(100, 6, "abcde"), (41, 40, "a" * 34), (57, 70, "z" * 68)])
def test_struct_return_1(get_contract, i, ln, s):
    contract = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]

@internal
def get_struct_x() -> X:
    return X(x={i}, y="{s}", z=b"{s}")

@external
def test() -> (int128, String[{ln}], Bytes[{ln}]):
    ret: X = self.get_struct_x()
    return ret.x, ret.y, ret.z
    """

    c = get_contract(contract)

    assert c.test() == (i, s, bytes(s, "utf-8"))


def test_dynamically_sized_struct_as_arg(get_contract):
    contract = """
struct X:
    x: uint256
    y: Bytes[6]

@internal
def _foo(x: X) -> Bytes[6]:
    return x.y

@external
def bar() -> Bytes[6]:
    _X: X = X(x=1, y=b"hello")
    return self._foo(_X)
    """

    c = get_contract(contract)

    assert c.bar() == b"hello"


def test_dynamically_sized_struct_as_arg_2(get_contract):
    contract = """
struct X:
    x: uint256
    y: String[6]

@internal
def _foo(x: X) -> String[6]:
    return x.y

@external
def bar() -> String[6]:
    _X: X = X(x=1, y="hello")
    return self._foo(_X)
    """

    c = get_contract(contract)

    assert c.bar() == "hello"


def test_dynamically_sized_struct_member_as_arg(get_contract):
    contract = """
struct X:
    x: uint256
    y: Bytes[6]

@internal
def _foo(s: Bytes[6]) -> Bytes[6]:
    return s

@external
def bar() -> Bytes[6]:
    _X: X = X(x=1, y=b"hello")
    return self._foo(_X.y)
    """

    c = get_contract(contract)

    assert c.bar() == b"hello"


def test_dynamically_sized_struct_member_as_arg_2(get_contract):
    contract = """
struct X:
    x: uint256
    y: String[6]

@internal
def _foo(s: String[6]) -> String[6]:
    return s

@external
def bar() -> String[6]:
    _X: X = X(x=1, y="hello")
    return self._foo(_X.y)
    """

    c = get_contract(contract)

    assert c.bar() == "hello"


# TODO probably want to refactor these into general test utils
st_uint256 = st.integers(min_value=0, max_value=2**256 - 1)
st_string65 = st.text(max_size=65, alphabet=string.printable)
st_bytes65 = st.binary(max_size=65)
st_sarray3 = st.lists(st_uint256, min_size=3, max_size=3)
st_darray3 = st.lists(st_uint256, max_size=3)

internal_call_kwargs_cases = [
    ("uint256", st_uint256),
    ("String[65]", st_string65),
    ("Bytes[65]", st_bytes65),
    ("uint256[3]", st_sarray3),
    ("DynArray[uint256, 3]", st_darray3),
]


@pytest.mark.parametrize("typ1,strategy1", internal_call_kwargs_cases)
@pytest.mark.parametrize("typ2,strategy2", internal_call_kwargs_cases)
def test_internal_call_kwargs(get_contract, typ1, strategy1, typ2, strategy2):
    # GHSA-ph9x-4vc9-m39g

    @given(kwarg1=strategy1, default1=strategy1, kwarg2=strategy2, default2=strategy2)
    @settings(max_examples=5)  # len(cases) * len(cases) * 5 * 5
    def fuzz(kwarg1, kwarg2, default1, default2):
        code = f"""
@internal
def foo(a: {typ1} = {repr(default1)}, b: {typ2} = {repr(default2)}) -> ({typ1}, {typ2}):
    return a, b

@external
def test0() -> ({typ1}, {typ2}):
    return self.foo()

@external
def test1() -> ({typ1}, {typ2}):
    return self.foo({repr(kwarg1)})

@external
def test2() -> ({typ1}, {typ2}):
    return self.foo({repr(kwarg1)}, {repr(kwarg2)})

@external
def test3(x1: {typ1}) -> ({typ1}, {typ2}):
    return self.foo(x1)

@external
def test4(x1: {typ1}, x2: {typ2}) -> ({typ1}, {typ2}):
    return self.foo(x1, x2)
        """
        c = get_contract(code)
        assert c.test0() == (default1, default2)
        assert c.test1() == (kwarg1, default2)
        assert c.test2() == (kwarg1, kwarg2)
        assert c.test3(kwarg1) == (kwarg1, default2)
        assert c.test4(kwarg1, kwarg2) == (kwarg1, kwarg2)

    fuzz()



from tests.evm_backends.base_env import _compile
from vyper.utils import method_id


def test_init_argument_test(get_contract):
    init_argument_test = """
moose: int128

@deploy
def __init__(_moose: int128):
    self.moose = _moose

@external
def returnMoose() -> int128:
    return self.moose
    """

    c = get_contract(init_argument_test, *[5])
    assert c.returnMoose() == 5
    print("Passed init argument test")


def test_constructor_mapping(get_contract):
    contract = """
foo: HashMap[bytes4, bool]

X: constant(bytes4) = 0x01ffc9a7

@deploy
def __init__():
    self.foo[X] = True

@external
@view
def check_foo(a: bytes4) -> bool:
    return self.foo[a]
    """

    c = get_contract(contract)
    assert c.check_foo("0x01ffc9a7") is True


def test_constructor_advanced_code(get_contract):
    constructor_advanced_code = """
twox: int128

@deploy
def __init__(x: int128):
    self.twox = x * 2

@external
def get_twox() -> int128:
    return self.twox
    """
    c = get_contract(constructor_advanced_code, *[5])
    assert c.get_twox() == 10


def test_constructor_advanced_code2(get_contract):
    constructor_advanced_code2 = """
comb: uint256

@deploy
def __init__(x: uint256[2], y: Bytes[3], z: uint256):
    self.comb = x[0] * 1000 + x[1] * 100 + len(y) * 10 + z

@external
def get_comb() -> uint256:
    return self.comb
    """
    c = get_contract(constructor_advanced_code2, *[[5, 7], b"dog", 8])
    assert c.get_comb() == 5738
    print("Passed advanced init argument tests")


def test_large_input_code(env, get_contract, tx_failed):
    large_input_code = """
@external
def foo(x: int128) -> int128:
    return 3
    """

    c = get_contract(large_input_code)
    signature = "(int128)"
    method = method_id(f"foo{signature}")

    env.message_call(c.address, data=method + (1274124).to_bytes(32, "big"))
    env.message_call(c.address, data=method + (2**120).to_bytes(32, "big"))

    with tx_failed():
        env.message_call(c.address, data=method + (2**130).to_bytes(32, "big"))


def test_large_input_code_2(env, get_contract, tx_failed, output_formats):
    code = """
@deploy
def __init__(x: int128):
    y: int128 = x

@external
def foo() -> int128:
    return 5
    """

    _, bytecode = _compile(code, output_formats)
    ctor_args = (2**127 - 1).to_bytes(32, "big")
    env._deploy(bytecode + ctor_args, value=0)

    with tx_failed():
        env._deploy(bytecode + (2**127).to_bytes(32, "big"), value=0)


def test_initialise_array_with_constant_key(get_contract):
    contract = """
X: constant(uint256) = 4

foo: int16[X]

@deploy
def __init__():
    self.foo[X-1] = -2

@external
@view
def check_foo(a: uint256) -> int16:
    return self.foo[a]
    """

    c = get_contract(contract)
    assert c.check_foo(3) == -2


def test_initialise_dynarray_with_constant_key(get_contract):
    contract = """
X: constant(int16) = 4

foo: DynArray[int16, X]

@deploy
def __init__():
    self.foo = [X - 3, X - 4, X - 5, X - 6]

@external
@view
def check_foo(a: uint64) -> int16:
    return self.foo[a]
    """

    c = get_contract(contract)
    assert c.check_foo(3) == -2


def test_nested_dynamic_array_constructor_arg(env, get_contract):
    code = """
foo: uint256

@deploy
def __init__(x: DynArray[DynArray[uint256, 3], 3]):
    self.foo = x[0][2] + x[1][1] + x[2][0]

@external
def get_foo() -> uint256:
    return self.foo
    """
    c = get_contract(code, *[[[3, 5, 7], [11, 13, 17], [19, 23, 29]]])
    assert c.get_foo() == 39


def test_nested_dynamic_array_constructor_arg_2(env, get_contract):
    code = """
foo: int128

@deploy
def __init__(x: DynArray[DynArray[DynArray[int128, 3], 3], 3]):
    self.foo = x[0][1][2] * x[1][1][1] * x[2][1][0] - x[0][0][0] - x[1][1][1] - x[2][2][2]

@external
def get_foo() -> int128:
    return self.foo
    """
    c = get_contract(
        code,
        *[
            [
                [[3, 5, 7], [11, 13, 17], [19, 23, 29]],
                [[-3, -5, -7], [-11, -13, -17], [-19, -23, -29]],
                [[-31, -37, -41], [-43, -47, -53], [-59, -61, -67]],
            ]
        ],
    )
    assert c.get_foo() == 9580


def test_initialise_nested_dynamic_array(env, get_contract):
    code = """
foo: DynArray[DynArray[uint256, 3], 3]

@deploy
def __init__(x: uint256, y: uint256, z: uint256):
    self.foo = [
        [x, y, z],
        [x * 1000 + y, y * 1000 + z, z * 1000 + x],
        [z * 2, y * 3, x * 4],
    ]

@external
def get_foo() -> DynArray[DynArray[uint256, 3], 3]:
    return self.foo
    """
    c = get_contract(code, *[37, 41, 73])
    assert c.get_foo() == [[37, 41, 73], [37041, 41073, 73037], [146, 123, 148]]


def test_initialise_nested_dynamic_array_2(env, get_contract):
    code = """
foo: DynArray[DynArray[DynArray[int128, 3], 3], 3]

@deploy
def __init__(x: int128, y: int128, z: int128):
    self.foo = [
        [[x, y, z], [y, z, x], [z, y, x]],
        [
            [x * 1000 + y, y * 1000 + z, z * 1000 + x],
            [- (x * 1000 + y), - (y * 1000 + z), - (z * 1000 + x)],
            [- (x * 1000) + y, - (y * 1000) + z, - (z * 1000) + x],
        ],
        [
            [z * 2, y * 3, x * 4],
            [z * (-2), y * (-3), x * (-4)],
            [z * (-y), y * (-x), x * (-z)],
        ],
    ]

@external
def get_foo() -> DynArray[DynArray[DynArray[int128, 3], 3], 3]:
    return self.foo
    """
    c = get_contract(code, *[37, 41, 73])
    assert c.get_foo() == [
        [[37, 41, 73], [41, 73, 37], [73, 41, 37]],
        [[37041, 41073, 73037], [-37041, -41073, -73037], [-36959, -40927, -72963]],
        [[146, 123, 148], [-146, -123, -148], [-2993, -1517, -2701]],
    ]



from vyper import ast as vy_ast


def test_compare_different_node_clases():
    vyper_ast = vy_ast.parse_to_ast("foo = 42")
    left = vyper_ast.body[0].target
    right = vyper_ast.body[0].value

    assert left != right


def test_compare_different_nodes_same_class():
    vyper_ast = vy_ast.parse_to_ast("[1, 2]")
    left, right = vyper_ast.body[0].value.elements

    assert left != right


def test_compare_different_nodes_same_value():
    vyper_ast = vy_ast.parse_to_ast("[1, 1]")
    left, right = vyper_ast.body[0].value.elements

    assert left != right


def test_compare_similar_node():
    # test equality without node_ids
    left = vy_ast.Int(value=1)
    right = vy_ast.Int(value=1)

    assert left == right


def test_compare_same_node():
    vyper_ast = vy_ast.parse_to_ast("42")
    node = vyper_ast.body[0].value

    assert node == node



def test_packing_test(get_contract):
    packing_test = """
struct Bar:
    a: int128
    b: int128
struct Z:
    foo: int128[3]
    bar: Bar[2]
x: int128
y: int128[5]
z: Z
a: int128

@external
def foo() -> int128:
    self.x = 1
    self.y[0] = 2
    self.y[4] = 4
    self.z.foo[0] = 8
    self.z.foo[2] = 16
    self.z.bar[0].a = 32
    self.z.bar[0].b = 64
    self.z.bar[1].a = 128
    self.z.bar[1].b = 256
    self.a = 512
    return self.x + self.y[0] + self.y[4] + self.z.foo[0] + self.z.foo[2] + \
        self.z.bar[0].a + self.z.bar[0].b + self.z.bar[1].a + self.z.bar[1].b + self.a

@external
def fop() -> int128:
    _x: int128 = 0
    _y: int128[5] = [0, 0, 0, 0, 0]
    _z: Z = Z(foo=[0, 0, 0], bar=[Bar(a=0, b=0), Bar(a=0, b=0)])
    _a: int128 = 0
    _x = 1
    _y[0] = 2
    _y[4] = 4
    _z.foo[0] = 8
    _z.foo[2] = 16
    _z.bar[0].a = 32
    _z.bar[0].b = 64
    _z.bar[1].a = 128
    _z.bar[1].b = 256
    _a = 512
    return _x + _y[0] + _y[4] + _z.foo[0] + _z.foo[2] + \
        _z.bar[0].a + _z.bar[0].b + _z.bar[1].a + _z.bar[1].b + _a
    """

    c = get_contract(packing_test)
    assert c.foo() == 1023, c.foo()
    assert c.fop() == 1023, c.fop()
    print("Passed packing test")



# Adapted from https://en.wikipedia.org/wiki/S-expression#Parsing


def parse_literal(word):
    try:
        return int(word)
    except ValueError:
        return word


def parse_s_exp(string):
    sexp = [[]]
    word = ""
    in_str = False
    in_comment = False
    for char in string:
        if in_comment:
            if char == "\n":  # comment ends at the end of a line
                in_comment = False
            continue
        if char == ";":  # start of comment
            in_comment = True
            continue
        if char == "(" and not in_str:
            sexp.append([])
        elif char == ")" and not in_str:
            if word:
                sexp[-1].append(parse_literal(word))
                word = ""
            temp = sexp.pop()
            sexp[-1].append(temp)
        elif char in (" ", "\n", "\t") and not in_str:
            if word:
                sexp[-1].append(parse_literal(word))
                word = ""
        elif char == '"':
            in_str = not in_str
        else:
            word += char
    return sexp[0]



import contextlib
import copy
import textwrap
import types

from vyper.compiler.settings import VYPER_ERROR_CONTEXT_LINES, VYPER_ERROR_LINE_NUMBERS


class ExceptionList(list):
    """
    List subclass for storing exceptions.
    To deliver multiple compilation errors to the user at once, append each
    raised Exception to this list and call raise_if_not_empty once the task
    is completed.
    """

    def raise_if_not_empty(self):
        if len(self) == 1:
            raise self[0]
        elif len(self) > 1:
            err_msg = ["Compilation failed with the following errors:"]
            err_msg += [f"{type(i).__name__}: {i}" for i in reversed(self)]
            raise VyperException("\n\n".join(err_msg))


class _BaseVyperException(Exception):
    """
    Base Vyper exception class.

    This exception is not raised directly. Other exceptions inherit it in
    order to display source annotations in the error string.
    """

    def __init__(self, message="Error Message not found.", *items, hint=None, prev_decl=None):
        """
        Exception initializer.

        Arguments
        ---------
        message : str
            Error message to display with the exception.
        *items : VyperNode | Tuple[str, VyperNode], optional
            Vyper ast node(s), or tuple of (description, node) indicating where
            the exception occurred. Source annotations are generated in the order
            the nodes are given.

            A single tuple of (lineno, col_offset) is also understood to support
            the old API, but new exceptions should not use this approach.
        """
        self._message = message
        self._hint = hint
        self.prev_decl = prev_decl

        self.lineno = None
        self.col_offset = None
        self.annotations = None

        if len(items) == 1 and isinstance(items[0], tuple) and isinstance(items[0][0], int):
            # support older exceptions that don't annotate - remove this in the future!
            self.lineno, self.col_offset = items[0][:2]
        else:
            # strip out None sources so that None can be passed as a valid
            # annotation (in case it is only available optionally)
            self.annotations = [k for k in items if k is not None]

    def with_annotation(self, *annotations):
        """
        Creates a copy of this exception with a modified source annotation.

        Arguments
        ---------
        *annotations : VyperNode | Tuple[str, VyperNode]
            AST node(s), or tuple of (description, node) to use in the annotation.

        Returns
        -------
        A copy of the exception with the new node offset(s) applied.
        """
        exc = copy.copy(self)
        exc.annotations = annotations
        return exc

    def append_annotation(self, exc):
        if self.annotations is None:
            self.annotations = []

        self.annotations = [exc] + self.annotations

    @property
    def hint(self):
        # some hints are expensive to compute, so we wait until the last
        # minute when the formatted message is actually requested to compute
        # them.
        if callable(self._hint):
            return self._hint()
        return self._hint

    @property
    def message(self):
        msg = self._message
        if self.hint:
            msg += f"\n\n  (hint: {self.hint})"
        return msg

    def format_annotation(self, value):
        from vyper import ast as vy_ast
        from vyper.utils import annotate_source_code

        node = value[1] if isinstance(value, tuple) else value
        node_msg = ""

        if isinstance(node, vy_ast.VyperNode):
            # folded AST nodes contain pointers to the original source
            node = node.get_original_node()

        try:
            source_annotation = annotate_source_code(
                # add trailing space because EOF exceptions point one char beyond the length
                f"{node.full_source_code} ",
                node.lineno,
                node.col_offset,
                context_lines=VYPER_ERROR_CONTEXT_LINES,
                line_numbers=VYPER_ERROR_LINE_NUMBERS,
            )
        except Exception:
            # necessary for certain types of syntax exceptions
            return None

        if isinstance(node, vy_ast.VyperNode):
            module_node = node.module_node

            # TODO: handle cases where module is None or vy_ast.Module
            if module_node.get("path") not in (None, "<unknown>"):
                node_msg = f'{node_msg}contract "{module_node.path}:{node.lineno}", '

            fn_node = node.get_ancestor(vy_ast.FunctionDef)
            if fn_node:
                node_msg = f'{node_msg}function "{fn_node.name}", '

        col_offset_str = "" if node.col_offset is None else str(node.col_offset)
        node_msg = f"{node_msg}line {node.lineno}:{col_offset_str} \n{source_annotation}\n"

        if isinstance(value, tuple):
            # if annotation includes a message, apply it at the start and further indent
            node_msg = textwrap.indent(node_msg, "  ")
            node_msg = f"{value[0]}\n{node_msg}"

        node_msg = textwrap.indent(node_msg, "  ")
        return node_msg

    def __str__(self):
        if not self.annotations:
            if self.lineno is not None and self.col_offset is not None:
                return f"line {self.lineno}:{self.col_offset} {self.message}"
            else:
                return self.message

        annotation_list = []

        if self.prev_decl is not None:
            formatted_decl = self.format_annotation(self.prev_decl)
            formatted_decl = f" (previously declared at):\n{formatted_decl}"
            annotation_list.append(formatted_decl)

        for value in self.annotations:
            annotation_list.append(self.format_annotation(value))

        annotation_list = [s for s in annotation_list if s is not None]
        annotation_msg = "\n".join(annotation_list)
        return f"{self.message}\n\n{annotation_msg}"


class VyperException(_BaseVyperException):
    pass


class SyntaxException(VyperException):

    """Invalid syntax."""

    def __init__(self, message, source_code, lineno, col_offset):
        item = types.SimpleNamespace()  # TODO: Create an actual object for this
        item.lineno = lineno
        item.col_offset = col_offset
        item.full_source_code = source_code
        super().__init__(message, item)


class DecimalOverrideException(VyperException):
    """The Vyper compiler uses specific Decimal settings which
    if overridden could lead to incorrect behavior.
    """


class NatSpecSyntaxException(SyntaxException):
    """Invalid syntax within NatSpec docstring."""


class StructureException(VyperException):
    """Invalid structure for parsable syntax."""


class InstantiationException(StructureException):
    """Variable or expression cannot be instantiated"""


class VersionException(SyntaxException):
    """Version string is malformed or incompatible with this compiler version."""


class VariableDeclarationException(VyperException):
    """Invalid variable declaration."""


class FunctionDeclarationException(VyperException):
    """Invalid function declaration."""


class FlagDeclarationException(VyperException):
    """Invalid flag declaration."""


class EventDeclarationException(VyperException):
    """Invalid event declaration."""


class UnknownType(VyperException):
    """Reference to a type that does not exist."""


class UnknownAttribute(VyperException):
    """Reference to an attribute that does not exist."""


class UndeclaredDefinition(VyperException):
    """Reference to a definition that has not been declared."""


class NamespaceCollision(VyperException):
    """Assignment to a name that is already in use."""


class InvalidLiteral(VyperException):
    """Invalid literal value."""


class InvalidAttribute(VyperException):
    """Reference to an attribute that does not exist."""


class InvalidReference(VyperException):
    """Invalid reference to an existing definition."""


class InvalidOperation(VyperException):
    """Invalid operator for a given type."""


class InvalidType(VyperException):
    """Type is invalid for an action."""


class TypeMismatch(VyperException):
    """Attempt to perform an action between multiple objects of incompatible types."""


class ArgumentException(VyperException):
    """Call to a function with invalid arguments."""


class CallViolation(VyperException):
    """Illegal function call."""


class ImportCycle(VyperException):
    """An import cycle"""


class DuplicateImport(VyperException):
    """A module was imported twice from the same module"""


class ModuleNotFound(VyperException):
    """Module was not found"""


class ImmutableViolation(VyperException):
    """Modifying an immutable variable, constant, or definition."""


class InitializerException(VyperException):
    """An issue with initializing/constructing a module"""


class BorrowException(VyperException):
    """An issue with borrowing/using a module"""


class StateAccessViolation(VyperException):
    """Violating the mutability of a function definition."""


class NonPayableViolation(VyperException):
    """msg.value in a nonpayable function."""


class InterfaceViolation(VyperException):
    """Interface is not fully implemented."""


class IteratorException(VyperException):
    """Improper use of iterators."""


class ArrayIndexException(VyperException):
    """Array index out of range."""


class ZeroDivisionException(VyperException):
    """Second argument to a division or modulo operation was zero."""


class OverflowException(VyperException):
    """Numeric value out of range for the given type."""


class EvmVersionException(VyperException):
    """Invalid action for the active EVM ruleset."""


class StorageLayoutException(VyperException):
    """Invalid slot for the storage layout overrides"""


class MemoryAllocationException(VyperException):
    """Tried to allocate too much memory"""


class JSONError(Exception):

    """Invalid compiler input JSON."""

    def __init__(self, msg, lineno=None, col_offset=None):
        super().__init__(msg)
        self.lineno = lineno
        self.col_offset = col_offset


class ParserException(Exception):
    """Contract source cannot be parsed."""


class BadArchive(Exception):
    """Bad archive"""


class UnimplementedException(VyperException):
    """Some feature is known to be not implemented"""


class FeatureException(VyperException):
    """Some feature flag is not enabled"""


class StaticAssertionException(VyperException):
    """An assertion is proven to fail at compile-time."""


class VyperInternalException(_BaseVyperException):
    """
    Base Vyper internal exception class.

    This exception is not raised directly, it is subclassed by other internal
    exceptions.

    Internal exceptions are raised as a means of telling the user that the
    compiler has panicked, and that filing a bug report would be appropriate.
    """

    def __str__(self):
        return (
            f"{super().__str__()}\n\n"
            "This is an unhandled internal compiler error. "
            "Please create an issue on Github to notify the developers!\n"
            "https://github.com/vyperlang/vyper/issues/new?template=bug.md"
        )


class CompilerPanic(VyperInternalException):
    """General unexpected error during compilation."""


class CodegenPanic(VyperInternalException):
    """Invalid code generated during codegen phase"""


class StackTooDeep(CodegenPanic):
    """Stack too deep"""  # (should not happen)


class UnexpectedNodeType(VyperInternalException):
    """Unexpected AST node type."""


class UnexpectedValue(VyperInternalException):
    """Unexpected Value."""


class UnfoldableNode(VyperInternalException):
    """Constant folding logic cannot be applied to an AST node."""


class TypeCheckFailure(VyperInternalException):
    """An issue was not caught during type checking that should have been."""


class InvalidABIType(VyperInternalException):
    """An internal routine constructed an invalid ABI type"""


@contextlib.contextmanager
def tag_exceptions(node, fallback_exception_type=CompilerPanic, note=None):
    try:
        yield
    except _BaseVyperException as e:
        if not e.annotations and not e.lineno:
            tb = e.__traceback__
            raise e.with_annotation(node).with_traceback(tb) from None
        raise e from None
    except Exception as e:
        tb = e.__traceback__
        fallback_message = f"unhandled exception {e}"
        if note:
            fallback_message += f", {note}"
        raise fallback_exception_type(fallback_message, node).with_traceback(tb)



import pytest

from vyper.evm.opcodes import version_check
from vyper.exceptions import StorageLayoutException

code = """

struct StructOne:
    a: String[33]
    b: uint256[3]

struct StructTwo:
    a: Bytes[5]
    b: int128[2]
    c: String[64]

a: public(StructOne)
b: public(uint256[2])
c: public(Bytes[32])
d: public(int128[4])
foo: public(HashMap[uint256, uint256[3]])
dyn_array: DynArray[uint256, 3]
e: public(String[48])
f: public(int256[1])
g: public(StructTwo[2])
h: public(int256[1])


@deploy
def __init__():
    self.a = StructOne(a="ok", b=[4,5,6])
    self.b = [7, 8]
    self.c = b"thisisthirtytwobytesokhowdoyoudo"
    self.d = [-1, -2, -3, -4]
    self.e = "A realllllly long string but we won't use it all"
    self.f = [33]
    self.g = [
        StructTwo(a=b"hello", b=[-66, 420], c="another string"),
        StructTwo(
            a=b"gbye",
            b=[1337, 888],
            c="whatifthisstringtakesuptheentirelengthwouldthatbesobadidothinkso"
        )
    ]
    self.dyn_array = [1, 2, 3]
    self.h =  [123456789]
    self.foo[0] = [987, 654, 321]
    self.foo[1] = [123, 456, 789]

@external
@nonreentrant
def with_lock():
    pass
"""


def test_storage_slots(get_contract):
    c = get_contract(code)
    assert c.a() == ("ok", [4, 5, 6])
    assert [c.b(i) for i in range(2)] == [7, 8]
    assert c.c() == b"thisisthirtytwobytesokhowdoyoudo"
    assert [c.d(i) for i in range(4)] == [-1, -2, -3, -4]
    assert c.e() == "A realllllly long string but we won't use it all"
    assert c.f(0) == 33
    assert c.g(0) == (b"hello", [-66, 420], "another string")
    assert c.g(1) == (
        b"gbye",
        [1337, 888],
        "whatifthisstringtakesuptheentirelengthwouldthatbesobadidothinkso",
    )
    assert [c.foo(0, i) for i in range(3)] == [987, 654, 321]
    assert [c.foo(1, i) for i in range(3)] == [123, 456, 789]
    assert c.h(0) == 123456789


def test_reentrancy_lock(get_contract):
    c = get_contract(code)

    # if re-entrancy locks are incorrectly placed within storage, these
    # calls will either revert or correupt the data that we read later
    c.with_lock()

    assert c.a() == ("ok", [4, 5, 6])
    assert [c.b(i) for i in range(2)] == [7, 8]
    assert c.c() == b"thisisthirtytwobytesokhowdoyoudo"
    assert [c.d(i) for i in range(4)] == [-1, -2, -3, -4]
    assert c.e() == "A realllllly long string but we won't use it all"
    assert c.f(0) == 33
    assert c.g(0) == (b"hello", [-66, 420], "another string")
    assert c.g(1) == (
        b"gbye",
        [1337, 888],
        "whatifthisstringtakesuptheentirelengthwouldthatbesobadidothinkso",
    )
    assert [c.foo(0, i) for i in range(3)] == [987, 654, 321]
    assert [c.foo(1, i) for i in range(3)] == [123, 456, 789]
    assert c.h(0) == 123456789


def test_allocator_overflow(get_contract):
    # cancun allocates reeentrancy slot in transient storage,
    # so allocate an actual storage variable
    if version_check(begin="cancun"):
        slot1 = "x: uint256"
    else:
        slot1 = "# --> global nonreentrancy slot allocated here <--"
    code = f"""
{slot1}
y: uint256[max_value(uint256)]
    """

    with pytest.raises(
        StorageLayoutException,
        match=f"Invalid storage slot, tried to allocate slots 1 through {2**256}",
    ):
        get_contract(code)



import random
import string

import hypothesis.strategies as st
import pytest
from hypothesis import given, settings

import vyper.ast as vy_ast
from vyper.ast.identifiers import RESERVED_KEYWORDS
from vyper.compiler.phases import CompilerData


def _valid_identifier(attr):
    return attr not in RESERVED_KEYWORDS


# random names for functions
@settings(max_examples=20)
@given(
    st.lists(
        st.tuples(
            st.sampled_from(["@pure", "@view", "@nonpayable", "@payable"]),
            st.text(alphabet=string.ascii_lowercase, min_size=1).filter(_valid_identifier),
        ),
        unique_by=lambda x: x[1],  # unique on function name
        min_size=1,
        max_size=10,
    )
)
@pytest.mark.fuzzing
def test_call_graph_stability_fuzz(funcs):
    def generate_func_def(mutability, func_name, i):
        return f"""
@internal
{mutability}
def {func_name}() -> uint256:
    return {i}
        """

    func_defs = "\n".join(generate_func_def(m, s, i) for i, (m, s) in enumerate(funcs))

    for _ in range(10):
        func_names = [f for (_, f) in funcs]
        random.shuffle(func_names)

        self_calls = "\n".join(f"  self.{f}()" for f in func_names)
        code = f"""
{func_defs}

@external
def foo():
{self_calls}
        """
        t = CompilerData(code)

        # check the .called_functions data structure on foo() directly
        foo = t.annotated_vyper_module.get_children(vy_ast.FunctionDef, filters={"name": "foo"})[0]
        foo_t = foo._metadata["func_type"]
        assert [f.name for f in foo_t.called_functions] == func_names

        # now for sanity, ensure the order that the function definitions appear
        # in the IR is the same as the order of the calls
        sigs = t.function_signatures
        del sigs["foo"]
        ir = t.ir_runtime
        ir_funcs = []
        # search for function labels
        for d in ir.args:  # currently: (seq ... (seq (label foo ...)) ...)
            if d.value == "seq" and d.args[0].value == "label":
                r = d.args[0].args[0].value
                if isinstance(r, str) and r.startswith("internal"):
                    ir_funcs.append(r)

        expected = []
        for f in foo_t.called_functions:
            expected.append(f._ir_info.internal_function_label(is_ctor_context=False))

        assert ir_funcs == expected



import hashlib

from vyper.utils import hex_to_int


def test_sha256_string_literal(get_contract):
    code = """
@external
def bar() -> bytes32:
    return sha256("test")
    """

    c = get_contract(code)

    assert c.bar() == hashlib.sha256(b"test").digest()


def test_sha256_literal_bytes(get_contract):
    code = """
@external
def bar() -> (bytes32 , bytes32):
    x: bytes32 = sha256("test")
    y: bytes32 = sha256(b"test")
    return x, y
    """
    c = get_contract(code)
    h = hashlib.sha256(b"test").digest()
    assert c.bar() == (h, h)


def test_sha256_bytes32(get_contract):
    code = """
@external
def bar(a: bytes32) -> bytes32:
    return sha256(a)
    """

    c = get_contract(code)

    test_val = 8 * b"bBaA"
    assert c.bar(test_val) == hashlib.sha256(test_val).digest()


def test_sha256_bytearraylike(get_contract):
    code = """
@external
def bar(a: String[100]) -> bytes32:
    return sha256(a)
    """

    c = get_contract(code)

    test_val = "test me! test me!"
    assert c.bar(test_val) == hashlib.sha256(test_val.encode()).digest()
    test_val = "fun"
    assert c.bar(test_val) == hashlib.sha256(test_val.encode()).digest()


def test_sha256_bytearraylike_storage(get_contract):
    code = """
a: public(Bytes[100])

@external
def set(b: Bytes[100]):
    self.a = b

@external
def bar() -> bytes32:
    return sha256(self.a)
    """

    c = get_contract(code)

    test_val = b"test me! test me!"
    c.set(test_val)
    assert c.a() == test_val
    assert c.bar() == hashlib.sha256(test_val).digest()


def test_sha256_constant_bytes32(get_contract):
    hex_val = "0x1234567890123456789012345678901234567890123456789012345678901234"
    code = f"""
FOO: constant(bytes32) = {hex_val}
BAR: constant(bytes32) = sha256(FOO)
@external
def foo() -> bytes32:
    x: bytes32 = BAR
    return x
    """
    c = get_contract(code)
    assert c.foo() == hashlib.sha256(hex_to_int(hex_val).to_bytes(32, "big")).digest()


def test_sha256_constant_string(get_contract):
    str_val = "0x1234567890123456789012345678901234567890123456789012345678901234"
    code = f"""
FOO: constant(String[66]) = "{str_val}"
BAR: constant(bytes32) = sha256(FOO)
@external
def foo() -> bytes32:
    x: bytes32 = BAR
    return x
    """
    c = get_contract(code)
    assert c.foo() == hashlib.sha256(str_val.encode()).digest()



import pytest

from vyper.exceptions import FunctionDeclarationException

# TODO test functions in this module across all evm versions
# once we have cancun support.


def test_nonreentrant_decorator(get_contract, tx_failed):
    malicious_code = """
interface ProtectedContract:
    def protected_function(callback_address: address): nonpayable

@external
def do_callback():
    extcall ProtectedContract(msg.sender).protected_function(self)
    """

    protected_code = """
interface Callbackable:
    def do_callback(): nonpayable

@external
@nonreentrant
def protected_function(c: Callbackable):
    extcall c.do_callback()

# add a default function so we know the callback didn't fail for any reason
# besides nonreentrancy
@external
def __default__():
    pass
    """
    contract = get_contract(protected_code)
    malicious = get_contract(malicious_code)

    with tx_failed():
        contract.protected_function(malicious.address)


def test_nonreentrant_view_function(get_contract, tx_failed):
    malicious_code = """
interface ProtectedContract:
    def protected_function(): nonpayable
    def protected_view_fn() -> uint256: view

@external
def do_callback() -> uint256:
    return staticcall ProtectedContract(msg.sender).protected_view_fn()
    """

    protected_code = """
interface Callbackable:
    def do_callback(): nonpayable

@external
@nonreentrant
def protected_function(c: Callbackable):
    extcall c.do_callback()

@external
@nonreentrant
@view
def protected_view_fn() -> uint256:
    return 10

# add a default function so we know the callback didn't fail for any reason
# besides nonreentrancy
@external
def __default__():
    pass
    """
    contract = get_contract(protected_code)
    malicious = get_contract(malicious_code)

    with tx_failed():
        contract.protected_function(malicious.address)


def test_multi_function_nonreentrant(get_contract, tx_failed):
    malicious_code = """
interface ProtectedContract:
    def unprotected_function(val: String[100], do_callback: bool): nonpayable
    def protected_function(val: String[100], do_callback: bool): nonpayable
    def special_value() -> String[100]: nonpayable

@external
def updated():
    extcall ProtectedContract(msg.sender).unprotected_function('surprise!', False)

@external
def updated_protected():
    # This should fail.
    extcall ProtectedContract(msg.sender).protected_function('surprise protected!', False)
    """

    protected_code = """
interface Callback:
    def updated(): nonpayable
    def updated_protected(): nonpayable

interface Self:
    def protected_function(val: String[100], do_callback: bool) -> uint256: nonpayable
    def protected_function2(val: String[100], do_callback: bool) -> uint256: nonpayable
    def protected_view_fn() -> String[100]: view

special_value: public(String[100])
callback: public(Callback)

@external
def set_callback(c: address):
    self.callback = Callback(c)

@external
@nonreentrant
def protected_function(val: String[100], do_callback: bool) -> uint256:
    self.special_value = val

    if do_callback:
        extcall self.callback.updated_protected()
        return 1
    else:
        return 2

@external
@nonreentrant
def protected_function2(val: String[100], do_callback: bool) -> uint256:
    self.special_value = val
    if do_callback:
        # call other function with same nonreentrancy key
        extcall Self(self).protected_function(val, False)
        return 1
    return 2

@external
@nonreentrant
def protected_function3(val: String[100], do_callback: bool) -> uint256:
    self.special_value = val
    if do_callback:
        # call other function with same nonreentrancy key
        assert self.special_value == staticcall Self(self).protected_view_fn()
        return 1
    return 2


@external
@nonreentrant
@view
def protected_view_fn() -> String[100]:
    return self.special_value

@external
def unprotected_function(val: String[100], do_callback: bool):
    self.special_value = val

    if do_callback:
        extcall self.callback.updated()

# add a default function so we know the callback didn't fail for any reason
# besides nonreentrancy
@external
def __default__():
    pass
    """
    contract = get_contract(protected_code)
    malicious = get_contract(malicious_code)

    contract.set_callback(malicious.address)
    assert contract.callback() == malicious.address

    # Test unprotected function.
    contract.unprotected_function("some value", True)
    assert contract.special_value() == "surprise!"

    # Test protected function.
    contract.protected_function("some value", False)
    assert contract.special_value() == "some value"
    assert contract.protected_view_fn() == "some value"

    with tx_failed():
        contract.protected_function("zzz value", True)

    contract.protected_function2("another value", False)
    assert contract.special_value() == "another value"

    with tx_failed():
        contract.protected_function2("zzz value", True)

    contract.protected_function3("another value", False)
    assert contract.special_value() == "another value"

    with tx_failed():
        contract.protected_function3("zzz value", True)


def test_nonreentrant_decorator_for_default(env, get_contract, tx_failed):
    calling_contract_code = """
@external
def send_funds(_amount: uint256):
    # raw_call() is used to overcome gas limit of send()
    response: Bytes[32] = raw_call(
        msg.sender,
        _abi_encode(msg.sender, _amount, method_id=method_id("transfer(address,uint256)")),
        max_outsize=32,
        value=_amount
    )

@external
@payable
def __default__():
    pass
    """

    reentrant_code = """
interface Callback:
    def send_funds(_amount: uint256): nonpayable

special_value: public(String[100])
callback: public(Callback)

@external
def set_callback(c: address):
    self.callback = Callback(c)

@external
@payable
@nonreentrant
def protected_function(val: String[100], do_callback: bool) -> uint256:
    self.special_value = val
    _amount: uint256 = msg.value
    send(self.callback.address, msg.value)

    if do_callback:
        extcall self.callback.send_funds(_amount)
        return 1
    else:
        return 2

@external
@payable
def unprotected_function(val: String[100], do_callback: bool):
    self.special_value = val
    _amount: uint256 = msg.value
    send(self.callback.address, msg.value)

    if do_callback:
        extcall self.callback.send_funds(_amount)

@external
@payable
@nonreentrant
def __default__():
    pass
    """

    reentrant_contract = get_contract(reentrant_code)
    calling_contract = get_contract(calling_contract_code)

    reentrant_contract.set_callback(calling_contract.address)
    assert reentrant_contract.callback() == calling_contract.address

    # Test unprotected function without callback.
    env.set_balance(env.deployer, 10**6)
    reentrant_contract.unprotected_function("some value", False, value=1000)
    assert reentrant_contract.special_value() == "some value"
    assert env.get_balance(reentrant_contract.address) == 0
    assert env.get_balance(calling_contract.address) == 1000

    # Test unprotected function with callback to default.
    reentrant_contract.unprotected_function("another value", True, value=1000)
    assert reentrant_contract.special_value() == "another value"
    assert env.get_balance(reentrant_contract.address) == 1000
    assert env.get_balance(calling_contract.address) == 1000

    # Test protected function without callback.
    reentrant_contract.protected_function("surprise!", False, value=1000)
    assert reentrant_contract.special_value() == "surprise!"
    assert env.get_balance(reentrant_contract.address) == 1000
    assert env.get_balance(calling_contract.address) == 2000

    # Test protected function with callback to default.
    with tx_failed():
        reentrant_contract.protected_function("zzz value", True, value=1000)


def test_disallow_on_init_function(get_contract):
    # nonreentrant has no effect when used on the __init__ fn
    # however, should disallow its usage regardless
    code = """

@external
@nonreentrant
def __init__():
    foo: uint256 = 0
"""
    with pytest.raises(FunctionDeclarationException):
        get_contract(code)



import pytest

from vyper import compiler

valid_list = [
    """
@internal
def mkint() -> int128:
    return 1

@external
def test_zerovalent():
    if True:
        self.mkint()

@external
def test_valency_mismatch():
    if True:
        self.mkint()
    else:
        pass
    """
]


@pytest.mark.parametrize("good_code", valid_list)
def test_conditional_return_code(good_code):
    assert compiler.compile_code(good_code) is not None



import copy
import warnings
from functools import cached_property
from pathlib import Path, PurePath
from typing import Optional

from vyper import ast as vy_ast
from vyper.ast import natspec
from vyper.codegen import module
from vyper.codegen.ir_node import IRnode
from vyper.compiler.input_bundle import FileInput, FilesystemInputBundle, InputBundle
from vyper.compiler.settings import OptimizationLevel, Settings, anchor_settings, merge_settings
from vyper.ir import compile_ir, optimizer
from vyper.semantics import analyze_module, set_data_positions, validate_compilation_target
from vyper.semantics.types.function import ContractFunctionT
from vyper.semantics.types.module import ModuleT
from vyper.typing import StorageLayout
from vyper.utils import ERC5202_PREFIX, vyper_warn
from vyper.venom import generate_assembly_experimental, generate_ir

DEFAULT_CONTRACT_PATH = PurePath("VyperContract.vy")


class CompilerData:
    """
    Object for fetching and storing compiler data for a Vyper contract.

    This object acts as a wrapper over the pure compiler functions, triggering
    compilation phases as needed and providing the data for use when generating
    the final compiler outputs.

    Attributes
    ----------
    vyper_module : vy_ast.Module
        Top-level Vyper AST node
    annotated_vyper_module: vy_ast.Module
        Annotated+analysed Vyper AST
    global_ctx : ModuleT
        Sorted, contextualized representation of the Vyper AST
    ir_nodes : IRnode
        IR used to generate deployment bytecode
    ir_runtime : IRnode
        IR used to generate runtime bytecode
    assembly : list
        Assembly instructions for deployment bytecode
    assembly_runtime : list
        Assembly instructions for runtime bytecode
    bytecode : bytes
        Deployment bytecode
    bytecode_runtime : bytes
        Runtime bytecode
    """

    def __init__(
        self,
        file_input: FileInput | str,
        input_bundle: InputBundle = None,
        settings: Settings = None,
        integrity_sum: str = None,
        storage_layout: StorageLayout = None,
        show_gas_estimates: bool = False,
        no_bytecode_metadata: bool = False,
    ) -> None:
        """
        Initialization method.

        Arguments
        ---------
        file_input: FileInput | str
            A FileInput or string representing the input to the compiler.
            FileInput is preferred, but `str` is accepted as a convenience
            method (and also for backwards compatibility reasons)
        settings: Settings
            Set optimization mode.
        show_gas_estimates: bool, optional
            Show gas estimates for abi and ir output modes
        no_bytecode_metadata: bool, optional
            Do not add metadata to bytecode. Defaults to False
        """

        if isinstance(file_input, str):
            file_input = FileInput(
                contents=file_input,
                source_id=-1,
                path=DEFAULT_CONTRACT_PATH,
                resolved_path=DEFAULT_CONTRACT_PATH,
            )
        self.file_input = file_input
        self.storage_layout_override = storage_layout
        self.show_gas_estimates = show_gas_estimates
        self.no_bytecode_metadata = no_bytecode_metadata
        self.original_settings = settings
        self.input_bundle = input_bundle or FilesystemInputBundle([Path(".")])
        self.expected_integrity_sum = integrity_sum

    @cached_property
    def source_code(self):
        return self.file_input.source_code

    @cached_property
    def source_id(self):
        return self.file_input.source_id

    @cached_property
    def contract_path(self):
        return self.file_input.path

    @cached_property
    def _generate_ast(self):
        settings, ast = vy_ast.parse_to_ast_with_settings(
            self.source_code,
            self.source_id,
            module_path=self.contract_path.as_posix(),
            resolved_path=self.file_input.resolved_path.as_posix(),
        )

        if self.original_settings:
            og_settings = self.original_settings
            settings = merge_settings(og_settings, settings)
            assert self.original_settings == og_settings  # be paranoid
        else:
            # merge with empty Settings(), doesn't do much but it does
            # remove the compiler version
            settings = merge_settings(Settings(), settings)

        if settings.optimize is None:
            settings.optimize = OptimizationLevel.default()

        if settings.experimental_codegen is None:
            settings.experimental_codegen = False

        return settings, ast

    @cached_property
    def settings(self):
        settings, _ = self._generate_ast
        return settings

    @cached_property
    def vyper_module(self):
        _, ast = self._generate_ast
        return ast

    @cached_property
    def _annotate(self) -> tuple[natspec.NatspecOutput, vy_ast.Module]:
        module = generate_annotated_ast(self.vyper_module, self.input_bundle)
        nspec = natspec.parse_natspec(module)
        return nspec, module

    @cached_property
    def natspec(self) -> natspec.NatspecOutput:
        return self._annotate[0]

    @cached_property
    def annotated_vyper_module(self) -> vy_ast.Module:
        return self._annotate[1]

    @cached_property
    def compilation_target(self):
        """
        Get the annotated AST, and additionally run the global checks
        required for a compilation target.
        """
        module_t = self.annotated_vyper_module._metadata["type"]

        expected = self.expected_integrity_sum

        if expected is not None and module_t.integrity_sum != expected:
            # warn for now. strict/relaxed mode was considered but it costs
            # interface and testing complexity to add another feature flag.
            vyper_warn(
                f"Mismatched integrity sum! Expected {expected}"
                f" but got {module_t.integrity_sum}."
                " (This likely indicates a corrupted archive)"
            )

        validate_compilation_target(module_t)
        return self.annotated_vyper_module

    @cached_property
    def storage_layout(self) -> StorageLayout:
        module_ast = self.compilation_target
        return set_data_positions(module_ast, self.storage_layout_override)

    @property
    def global_ctx(self) -> ModuleT:
        # ensure storage layout is computed
        _ = self.storage_layout
        # ensure natspec is computed
        _ = self.natspec
        return self.annotated_vyper_module._metadata["type"]

    @cached_property
    def _ir_output(self):
        # fetch both deployment and runtime IR
        return generate_ir_nodes(self.global_ctx, self.settings)

    @property
    def ir_nodes(self) -> IRnode:
        ir, ir_runtime = self._ir_output
        return ir

    @property
    def ir_runtime(self) -> IRnode:
        ir, ir_runtime = self._ir_output
        return ir_runtime

    @property
    def function_signatures(self) -> dict[str, ContractFunctionT]:
        # some metadata gets calculated during codegen, so
        # ensure codegen is run:
        _ = self._ir_output

        fs = self.annotated_vyper_module.get_children(vy_ast.FunctionDef)
        return {f.name: f._metadata["func_type"] for f in fs}

    @cached_property
    def venom_functions(self):
        deploy_ir, runtime_ir = self._ir_output
        deploy_venom = generate_ir(deploy_ir, self.settings.optimize)
        runtime_venom = generate_ir(runtime_ir, self.settings.optimize)
        return deploy_venom, runtime_venom

    @cached_property
    def assembly(self) -> list:
        if self.settings.experimental_codegen:
            deploy_code, runtime_code = self.venom_functions
            assert self.settings.optimize is not None  # mypy hint
            return generate_assembly_experimental(
                runtime_code, deploy_code=deploy_code, optimize=self.settings.optimize
            )
        else:
            return generate_assembly(self.ir_nodes, self.settings.optimize)

    @cached_property
    def assembly_runtime(self) -> list:
        if self.settings.experimental_codegen:
            _, runtime_code = self.venom_functions
            assert self.settings.optimize is not None  # mypy hint
            return generate_assembly_experimental(runtime_code, optimize=self.settings.optimize)
        else:
            return generate_assembly(self.ir_runtime, self.settings.optimize)

    @cached_property
    def bytecode(self) -> bytes:
        insert_compiler_metadata = not self.no_bytecode_metadata
        return generate_bytecode(self.assembly, insert_compiler_metadata=insert_compiler_metadata)

    @cached_property
    def bytecode_runtime(self) -> bytes:
        return generate_bytecode(self.assembly_runtime, insert_compiler_metadata=False)

    @cached_property
    def blueprint_bytecode(self) -> bytes:
        blueprint_bytecode = ERC5202_PREFIX + self.bytecode

        # the length of the deployed code in bytes
        len_bytes = len(blueprint_bytecode).to_bytes(2, "big")
        deploy_bytecode = b"\x61" + len_bytes + b"\x3d\x81\x60\x0a\x3d\x39\xf3"

        return deploy_bytecode + blueprint_bytecode


def generate_annotated_ast(vyper_module: vy_ast.Module, input_bundle: InputBundle) -> vy_ast.Module:
    """
    Validates and annotates the Vyper AST.

    Arguments
    ---------
    vyper_module : vy_ast.Module
        Top-level Vyper AST node

    Returns
    -------
    vy_ast.Module
        Annotated Vyper AST
    """
    vyper_module = copy.deepcopy(vyper_module)
    with input_bundle.search_path(Path(vyper_module.resolved_path).parent):
        # note: analyze_module does type inference on the AST
        analyze_module(vyper_module, input_bundle)

    return vyper_module


def generate_ir_nodes(global_ctx: ModuleT, settings: Settings) -> tuple[IRnode, IRnode]:
    """
    Generate the intermediate representation (IR) from the contextualized AST.

    This phase also includes IR-level optimizations.

    This function returns three values: deployment bytecode, runtime bytecode
    and the function signatures of the contract

    Arguments
    ---------
    global_ctx: ModuleT
        Contextualized Vyper AST

    Returns
    -------
    (IRnode, IRnode)
        IR to generate deployment bytecode
        IR to generate runtime bytecode
    """
    with anchor_settings(settings):
        ir_nodes, ir_runtime = module.generate_ir_for_module(global_ctx)
    if settings.optimize != OptimizationLevel.NONE:
        ir_nodes = optimizer.optimize(ir_nodes)
        ir_runtime = optimizer.optimize(ir_runtime)
    return ir_nodes, ir_runtime


def generate_assembly(ir_nodes: IRnode, optimize: Optional[OptimizationLevel] = None) -> list:
    """
    Generate assembly instructions from IR.

    Arguments
    ---------
    ir_nodes : str
        Top-level IR nodes. Can be deployment or runtime IR.

    Returns
    -------
    list
        List of assembly instructions.
    """
    optimize = optimize or OptimizationLevel.default()
    assembly = compile_ir.compile_to_assembly(ir_nodes, optimize=optimize)

    if _find_nested_opcode(assembly, "DEBUG"):
        warnings.warn(
            "This code contains DEBUG opcodes! The DEBUG opcode will only work in "
            "a supported EVM! It will FAIL on all other nodes!",
            stacklevel=2,
        )
    return assembly


def _find_nested_opcode(assembly, key):
    if key in assembly:
        return True
    else:
        sublists = [sub for sub in assembly if isinstance(sub, list)]
        return any(_find_nested_opcode(x, key) for x in sublists)


def generate_bytecode(assembly: list, insert_compiler_metadata: bool) -> bytes:
    """
    Generate bytecode from assembly instructions.

    Arguments
    ---------
    assembly : list
        Assembly instructions. Can be deployment or runtime assembly.

    Returns
    -------
    bytes
        Final compiled bytecode.
    """
    return compile_ir.assembly_to_evm(assembly, insert_compiler_metadata=insert_compiler_metadata)[
        0
    ]



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import NonPayableViolation

fail_list = [
    """
@external
def foo():
    x: uint256 = msg.value
"""
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_variable_declaration_exception(bad_code):
    with raises(NonPayableViolation):
        compiler.compile_code(bad_code)


valid_list = [
    """
x: int128
@external
@payable
def foo() -> int128:
    self.x = 5
    return self.x
    """,
    """
@external
@payable
def foo():
    x: uint256 = msg.value
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compiler.compile_code(good_code) is not None
